<?php
/*
Plugin Name: WordPress Migration
Plugin URI: http://gsoc2010.wordpress.com/brian-mckenna-automatic-migration/
Description: Migrate your blog from this URL to another URL.
Author: Brian McKenna
Author URI: http://brianmckenna.org/
*/

class WP_Migration {

	// XML-RPC client
	var $client;

	// Helpful status messages from the server
	var $responses = array();

        function WP_Migration() {
		$this->__construct();
        }

        function __construct() {
		add_action('admin_menu', array(&$this, 'adminMenu'));
		add_filter('xmlrpc_methods', array(&$this, 'xmlrpcMethods'));
		add_action('wp_ajax_migrate_files', array(&$this, 'ajaxFiles'));
		add_action('wp_ajax_migrate_data', array(&$this, 'ajaxData'));
	}

	function xmlrpcMethods($methods) {
		$methods['migration.setup'] = array(&$this, 'setup');
		$methods['migration.makePost'] = array(&$this, 'makePost');
		$methods['migration.makePostMeta'] = array(&$this, 'makePostMeta');
		$methods['migration.makeUser'] = array(&$this, 'makeUser');
		$methods['migration.makeComment'] = array(&$this, 'makeComment');
		$methods['migration.makeTaxonomy'] = array(&$this, 'makeTaxonomy');
		$methods['migration.makeLinks'] = array(&$this, 'makeLinks');
		$methods['migration.updateOptions'] = array(&$this, 'updateOptions');
		$methods['migration.finish'] = array(&$this, 'finish');
		return $methods;
	}

	function translateURL($data) {
		$options = get_option('migration_options');
		$from_url = $options['siteurl'];
		$to_url = get_option('siteurl');
		return str_replace($from_url, $to_url, $data);
	}

	function filesystemMethod($method) {
		// FIXME: Might not have sockets nor ftp extensions installed
		if($method == 'direct') {
			return 'ftpsockets';
		}

		return $method;
	}

	/**
	 *
	 *
	 * @param array $args Method Parameters.
	 * @return string
	 */
	function setup($args) {
		add_option('migration_options', $args);

		// Delete all posts
		$query = new WP_Query(array(
			'post_type' => 'any',
			'post_status' => 'any',
		));
		$deleted = array();
		foreach($query->posts as $post) {
			$deleted[] = $post->ID;
			wp_delete_post($post->ID, true);
		}

		// Delete all users
		$query = new WP_User_Search();
		foreach($query->get_results() as $userid) {
			wp_delete_user($userid, $userid);
		}

		// Delete all links
		$links = get_bookmarks();
		foreach($links as $link) {
			wp_delete_link($link->link_id);
		}

		return _('Started migration');
	}

	/**
	 *
	 *
	 * @param array $args Method Parameters.
	 * @return string
	 */
	function finish($args) {
		delete_option('migration_options');
		return _('Finished migration');
	}

	/**
	 * Makes a post from the passed in data.
	 *
	 * @param array $args Method Parameters.
	 * @return string
	 */
	function makePost($args) {
		global $wpdb;

		$postdata = $args;
		
		$post_ID = $postdata['ID'];

		$postdata['ID'] = NULL;
		$postdata['post_content'] = $this->translateURL($postdata['post_content']);
		$postdata['post_excerpt'] = $this->translateURL($postdata['post_excerpt']);
		$postdata['guid'] = $this->translateURL($postdata['guid']);

		$new_ID = wp_insert_post($postdata, true);

		// Update to the given ID
		$wpdb->update($wpdb->posts, array('ID' => $post_ID), array('ID' => $new_ID));
		$wpdb->update($wpdb->postmeta, array('post_id' => $post_ID), array('post_id' => $new_ID));

		return _('Saved post #') . $post_ID;
	}

	/**
	 *
	 *
	 * @param array $args Method Parameters.
	 * @return string
	 */
	function makePostMeta($args) {
		$post_ID = $args[0];

		// Out with the old meta
		$old_meta = get_post_meta($post_ID, NULL);
		foreach($old_meta as $meta_key => $values) {
			delete_post_meta($post_ID, $meta_key);
		}

		// In with the new meta
		$all_meta = $args[1];
		foreach($all_meta as $meta_key => $values) {
			foreach($values as $value) {
				$value = maybe_unserialize($this->translateURL($value));
				add_post_meta($post_ID, $meta_key, $value);
			}
		}

		return _('Saved meta for post #') . $post_ID;
	}

	/**
	 * Makes a user from the passed in data.
	 *
	 * @param array $args Method Parameters.
	 * @return string
	 */
	function makeUser($args) {
		global $wpdb;

		$user = $args[0];
		$userdata = $user['data'];
		$user_ID = $userdata['ID'];

		// Make a new user
		$userdata['ID'] = NULL;
		$new_ID = wp_insert_user($userdata, true);

		// Update to the given ID
		$wpdb->update($wpdb->users, array('ID' => $user_ID), array('ID' => $new_ID));
		$wpdb->update($wpdb->usermeta, array('user_id' => $user_ID), array('user_id' => $new_ID));

		// Make sure the password is the same
		$wpdb->update($wpdb->users, array('user_pass' => $userdata['user_pass']), array('ID' => $user_ID));

		// Out with the old meta
		$old_meta = get_user_meta($user_ID, NULL);
		foreach($old_meta as $meta_key => $values) {
			delete_user_meta($user_ID, $meta_key);
		}

		// In with the new meta
		$all_meta = $args[1];
		foreach($all_meta as $meta_key => $values) {
			foreach($values as $value) {
				$value = maybe_unserialize($this->translateURL($value));
				add_user_meta($user_ID, $meta_key, $value);
			}
		}

		return _('Saved user #') . $user_ID;
	}

	/**
	 *
	 *
	 * @param array $args Method Parameters.
	 * @return string
	 */
	function makeComment($args) {
		$commentdata = $args;
		$comment_ID = wp_insert_comment($commentdata);
		return _('Saved comment #') . $comment_ID;
	}

	/**
	 *
	 *
	 * @param array $args Method Parameters.
	 * @return string
	 */
	function makeTaxonomy($args) {
		$taxonomy = $args[0];
		register_taxonomy($taxonomy['name'], $taxonomy['object_type'], $taxonomy);

		$terms = $args[1];
		foreach($terms as $term) {
			wp_insert_term($term['name'], $taxonomy['name'], $term);
		}

		return _('Saved taxonomy: ') . $taxonomy['name'];
	}

	/**
	 *
	 *
	 * @param array $args Method Parameters.
	 * @return string
	 */
	function makeLinks($args) {
		$links = $args;

		foreach($links as $link) {
			$link['link_id'] = NULL;
			wp_insert_link($link);
		}

		return _('Saved links');
	}

	/**
	 *
	 *
	 * @param array $args Method Parameters.
	 * @return string
	 */
	function updateOptions($args) {
		$options = $args;
		foreach($options as $name => $value) {
			$value = maybe_unserialize($this->translateURL($value));
			update_option($name, $value);
		}

		return _('Copied site options');
	}

        function adminMenu() {
		$page = add_options_page(__('Migrate', 'migrate'), __('Migrate', 'migrate'), 8, 'wp-migration', array(&$this, 'adminPage'));
		add_action('admin_print_scripts-' . $page, array(&$this, 'adminPrintScripts'));
		add_filter('filesystem_method', array(&$this, 'filesystemMethod'));
        }

        function adminPrintScripts() {
		wp_enqueue_script('migration-ajax', plugins_url('migration-ajax.js', __FILE__), array('jquery'));
		if($_GET['page'] = 'wp-migration' && $_GET['step'] == 'migrate') {
			$siteurl = get_option('migration_url');
			$initurl = untrailingslashit($siteurl) . '/wp-content/plugins/wp-migration/init.php';

			wp_localize_script('migration-ajax', 'migrationSettings', array(
				'initurl' => $initurl,
			));
		}
	}

	function transferFiles($files) {
		global $wp_filesystem;

		$success_count = 0;
		$exists_count = 0;
		$error_count = 0;

		foreach($files as $fullpath => $file) {
			if($wp_filesystem->exists($fullpath)) {
				$exists_count++;
				continue;
			}

			if($file['type'] == 'd') {
				$wp_filesystem->mkdir($fullpath);
				continue;
			}

			$dir = dirname($fullpath);
			if(!$wp_filesystem->exists($dir)) {
				$wp_filesystem->mkdir($dir);
			}
			
			$success = $wp_filesystem->copy(ABSPATH . $fullpath, $fullpath);
			if($success) {
				$success_count++;
			} else {
				$error_count++;
			}
		}

		return array(
			'successes' => $success_count,
			'exists' => $exists_count,
			'errors' => $error_count,
		);
	}

	function flattenTree($files, $path = '') {
		$flattened = array();

		foreach($files as $file) {
			$dir = trailingslashit($path);
			if(empty($path)) {
				$dir = '';
			}
			$fullpath = $dir . $file['name'];

			// Only keep the info that we need
			$flattened[$fullpath] = array(
				'type' => $file['type'],
				'size' => $file['size'],
			);

			if($file['type'] == 'd') {
				$children = $this->flattenTree($file['files'], $fullpath);
				foreach($children as $fullpath => $child) {
					$flattened[$fullpath] = $child;
				}
			}
		}

		if(empty($path)) {
			// We don't want the wp-config.php file copied over!
			unset($flattened['wp-config.php']);
		}

		return $flattened;
	}

	function setupFiles() {
		require_once ABSPATH . 'wp-admin/includes/class-wp-filesystem-base.php';
		require_once ABSPATH . 'wp-admin/includes/class-wp-filesystem-direct.php';

		$direct = new WP_Filesystem_Direct(null);
		$tree = $direct->dirlist(ABSPATH, true, true);
		$files = $this->flattenTree($tree);
		update_option('migration_files', $files);
	}

	function ajaxFiles() {
		global $wp_filesystem;
		require_once ABSPATH . 'wp-admin/includes/file.php';

		$path = get_option('migration_path');

		$credentials = get_option('migration_credentials');
		WP_Filesystem($credentials);
		$wp_filesystem->chdir($path);

		$files = get_option('migration_files');
		if(empty($files)) {
			// Lastly, transfer over the new wp-config.php
			$wpconfig = get_option('migration_wpconfig');
			$wp_filesystem->put_contents('wp-config.php', $wpconfig);

			die('-1');
		}

		// Keep selecting files to transfer below a threshold
		$max_kbytes = 175;

		$top_files = array();
		$top_files_kbytes = 0;
		foreach($files as $fullpath => $file) {
			$file_kbytes = $file['size'] / 1024;

			if(count($top_files) > 1 && ($top_files_kbytes + $file_kbytes) > $max_kbytes) break;

			$top_files_kbytes += $file_kbytes;
			$top_files[$fullpath] = $file;
		}
		array_splice($files, 0, count($top_files));

		$info = $this->transferFiles($top_files);

		// We've transferred the top 10 files, remove them from the options
		update_option('migration_files', $files);

		$message = sprintf(__('Transferred %d files. %d files already existed. %d errors.'), $info['successes'], $info['exists'], $info['errors']);
		$json_message = json_encode($message);
		die($json_message);
	}

	function ajaxData() {
		$location = untrailingslashit(get_option('migration_url')) . '/xmlrpc.php';

		include_once(ABSPATH . WPINC . '/class-IXR.php');

		$this->client = new IXR_Client($location);
		$this->sendSetup();
		$this->sendPosts();
		$this->sendUsers();
		$this->sendComments();
		$this->sendTaxonomies();
		$this->sendLinks();
		$this->sendOptions();
		$this->sendFinish();

		$json_response = json_encode($this->responses);
		die($json_response);
	}

	function respond($message) {
		if($this->client->isError()) {
			$this->responses[] = $this->client->getErrorMessage();
		} else {
			$this->responses[] = $this->client->getResponse();
		}
	}

	function sendPosts() {
		$query = new WP_Query(array(
			'post_type' => 'any',
			'post_status' => 'any',
		));
		foreach($query->posts as $post) {
			$this->client->query('migration.makePost', $post);
			$this->respond();

			$all_meta = get_post_meta($post->ID, NULL);
			$this->client->query('migration.makePostMeta', $post->ID, $all_meta);
			$this->respond();
		}
	}

	function sendUsers() {
		$query = new WP_User_Search();
		foreach($query->get_results() as $userid) {
			$user = new WP_User($userid);
			$all_meta = get_user_meta($userid, NULL);

			$this->client->query('migration.makeUser', $user, $all_meta);
			$this->respond();
		}
	}

	function sendComments() {
		$comments = get_comments();
		foreach($comments as $comment) {
			$this->client->query('migration.makeComment', $comment);
			$this->respond();
		}
	}

	function sendTaxonomies() {
		$taxonomies = get_taxonomies();

		foreach($taxonomies as $taxonomy_name) {
			$taxonomy = get_taxonomy($taxonomy_name);
			$terms = get_terms($taxonomy_name, array('hide_empty' => false));

			$this->client->query('migration.makeTaxonomy', $taxonomy, $terms);
			$this->respond();
		}
	}

	function sendLinks() {
		$links = get_bookmarks();

		$this->client->query('migration.makeLinks', $links);
		$this->respond();
	}

	function sendOptions() {
		$alloptions = wp_load_alloptions();

		$this->client->query('migration.updateOptions', $alloptions);
		$this->respond();
	}

	function sendSetup() {
		$options = array(
			'siteurl' => get_option('siteurl'),
		);
		$this->client->query('migration.setup', $options);
		$this->respond();
	}

	function sendFinish() {
		$this->client->query('migration.finish');
		$this->respond();
	}

	function requestPath() {
		global $wp_filesystem;

		$credentials = get_option('migration_credentials');
		WP_Filesystem($credentials);

		$base = '/';
		if(isset($_POST['base']) && isset($_POST['dir'])) {
			if($_POST['base'] == $base) {
				$_POST['base'] = '';
			}

			$base = $_POST['base'] . '/' . $_POST['dir'];
		}

		if(isset($_POST['path'])) {
			return $_POST['base'];
		}

		$url = add_query_arg(array('step' => 'setup'));
?>
<form action="<?php echo esc_attr($url); ?>" method="post">
  <p>Current dir is <?php echo esc_attr($base); ?> <input type="submit" name="path" value="Use this path" /></p>
  <input type="hidden" name="base" value="<?php echo esc_attr($base); ?>" />
  <ul>
<?php
		$list = $wp_filesystem->dirlist($base);
		foreach($list as $file) {
			if($file['isdir']) {
?>
    <li><input type="submit" name="dir" value="<?php echo esc_attr($file['name']); ?>" /></li>
<?php
			}
		}
?>
  </ul>
</form>
<?php
	}

	function requestDetails() {
		if(isset($_POST['migration_url']) && isset($_POST['migration_wpconfig'])) {
			return array(
				'url' => $_POST['migration_url'],
				'wpconfig' => stripslashes($_POST['migration_wpconfig'])
			);
		}

		$wpconfig = file_get_contents(ABSPATH . 'wp-config.php');

		$url = add_query_arg(array('step' => 'migrate'));
?>
<form action="<?php echo esc_attr($url); ?>" method="post">
  <h3>New settings</h3>
  <table class="form-table">
    <tr>
      <th><label for="migration_url">URL</label></th>
      <td><input type="text" id="migration_url" name="migration_url" value="" /></td>
    </tr>
    <tr>
      <th>wp-config.php</th>
      <td><textarea id="migration_wpconfig" name="migration_wpconfig" rows="25" cols="70"><?php echo esc_attr($wpconfig); ?></textarea></td>
    </tr>
  </table>
  <p>
  </p>
  <p class="submit">
    <input class="button-primary" type="submit" value="Migrate" />
  </p>
</form>
<?php
	}

	function getCredentials($use_option = true) {
		$url = add_query_arg(array('step' => 'setup'));

		$credentials = get_option('migration_credentials');
		if(empty($credentials)) $credentials = request_filesystem_credentials($url);
		if(empty($credentials)) return false;
		if(!WP_Filesystem($credentials)) return request_filesystem_credentials($url, '', true);

		update_option('migration_credentials', $credentials);
		return true;
	}

	function getPath() {
		$path = get_option('migration_path');
		if(empty($path)) $path = $this->requestPath();
		if(empty($path)) return false;

		update_option('migration_path', $path);
		return true;
	}

	function getDetails() {
		$details = $this->requestDetails();
		if(empty($details)) return false;

		update_option('migration_url', $details['url']);
		update_option('migration_wpconfig', $details['wpconfig']);
		return true;
	}

	function deleteOptions() {
		delete_option('migration_credentials');

		delete_option('migration_path');

		delete_option('migration_url');
		delete_option('migration_wpconfig');
	}

	function migrate() {
		$this->setupFiles();
?>
<div class="wrap">
  <?php screen_icon(); ?>
  <h2><?php _e( 'Migrate Installation' ); ?></h2>

  <h3>Responses</h3>
  <ul id="server-response-list"></ul>
  <div id="server-response-loading" class="loading">Loading...</div>
</div>
<?php
	}

	function adminPage() {
		$setup = 'default';
		if(isset($_GET['step'])) {
			$step = $_GET['step'];
		}

		switch($step) {
		case 'setup':
			if($this->getCredentials()) {
				if($this->getPath()) {
					$this->getDetails();
				}
			}
			break;
		case 'migrate':
			if($this->getDetails()) {
				$this->migrate();
			}
			break;
		default:
			$this->deleteOptions();
			$this->getCredentials();
			break;
		}
	}
};

new WP_Migration();
?>
